unit uTrueType;
{.DEFINE TESTCHECKSUMS}
{$DEFINE UPDATECHECKSUMS}

interface

uses
  SysUtils, Classes,
  uTrueTypeStruct, uTrueTypeConst, uTrueTypeFunc,
  uTrueTypeInternalStruct,
  uStringUtils;

type
  ETrueTypeError = class(Exception);

  TTrueTypeFont = class(TComponent)
    private
      fValid: boolean;
      fOffsetTable: TT_OffsetTable;
      fTableDirectory: array of TT_TableDirectory;
      fTables: array of pointer;
      fFontNaming: TFontNameArray;
      fFontNamingChanged: boolean;
    protected
      procedure CheckValidity; virtual;
      procedure CheckTableValidity(Index: integer); virtual;
      function CalculateTableChecksum(Index: integer): LongWord; virtual;
      function VerifyTableChecksum(Index: integer): boolean; virtual;
      procedure UpdateTableChecksum(Index: integer); virtual;
      function FindTableIndex(Tag: ShortString; StartingIndex: integer = 0): integer; virtual;
      procedure UpdateTable(Index: integer; const Data; const Size: LongWord); virtual;
      procedure ApplyNamingRecordChanges; virtual;
      function IsFontNamingUnicode(NamingRecord: PTT_NamingTableNameRecord): boolean; overload; virtual;
      function IsFontNamingUnicode(PlatformID, EncodingID: integer): boolean; overload; virtual;
      procedure ReadFontNamingTable; virtual;
      procedure BuildFontNamingTable(out Table: PTT_NamingTable; out TableSize: integer); virtual;
    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      procedure Clear; virtual;
      procedure ApplyChanges; virtual;
      procedure CancelChanges; virtual;
      procedure LoadFromStream(Stream: TStream); virtual;
      procedure LoadFromFile(const FileName: string); virtual;
      procedure SaveToStream(Stream: TStream); virtual;
      procedure SaveToFile(const FileName: string); virtual;
      function GetAllFontNaming(FontNameID: integer; out Names: TWideStringArray; PreferredPlatform: integer = -1; PreferredEncoding: integer = -1; PreferredLanguage: integer = -1): WideString; virtual;
      function GetFontNaming(FontNameID: integer; PreferredPlatform: integer = -1; PreferredEncoding: integer = -1; PreferredLanguage: integer = -1): WideString; virtual;
      procedure SetFontNaming(FontNameID: integer; const Value: WideString; PreferredPlatform: integer = -1; PreferredEncoding: integer = -1; PreferredLanguage: integer = -1); virtual;
    published
    end;

implementation

{ TTrueTypeFont }

constructor TTrueTypeFont.Create(AOwner: TComponent);
begin
  inherited;
  Clear;
end;

destructor TTrueTypeFont.Destroy;
begin
  Clear;
  inherited;
end;

procedure TTrueTypeFont.Clear;
var i: integer;
begin
  fValid := False;
  FillChar(fOffsetTable, sizeof(fOffsetTable), 0);
  SetLength(fTableDirectory, 0);
  for i := 0 to Pred(Length(fTables)) do
    if fTables[i] <> nil then
      FreeMem(fTables[i]);
  SetLength(fTables, 0);
  SetLength(fFontNaming, 0);
  fFontNamingChanged := False;
end;

procedure TTrueTypeFont.CheckValidity;
begin
  if not fValid then
    Raise ETrueTypeError.Create('No valid font loaded.');
end;

procedure TTrueTypeFont.CheckTableValidity(Index: integer);
begin
  CheckValidity;
  if Index >= TTUShortToInt(fOffsetTable.numTables) then
    Raise ETrueTypeError.Create('Table index out of range.');
end;

procedure TTrueTypeFont.LoadFromStream(Stream: TStream);

  procedure CheckTable(const TableTag: AnsiString);
    begin
      if FindTableIndex(TableTag) < 0 then
        Raise ETrueTypeError.Create('Failed to locate table "' + TableTag + '"');
    end;

var i, n: integer;
    sz: LongWord;
begin
  Clear;
  Stream.ReadBuffer(fOffsetTable, Sizeof(fOffsetTable));
  n := TTUShortToInt(fOffsetTable.numTables);
  SetLength(fTableDirectory, n);
  FillChar(fTableDirectory, 0, sizeof(fTableDirectory));
  for i := 0 to Pred(n) do
    Stream.ReadBuffer(fTableDirectory[i], Sizeof(fTableDirectory[i]));
  SetLength(fTables, n);
  FillChar(fTables, 0, sizeof(fTables));
  for i := 0 to Pred(n) do
    begin
    sz := TTUlongToInt(fTableDirectory[i].length);
    GetMem(fTables[i], sz);
    Stream.Position := TTULongToInt(fTableDirectory[i].offset);
    Stream.ReadBuffer(fTables[i]^, sz);
    end;
  fValid := True;
  try
    CheckTable(TT_TAG_CHARACTERTOGLYPHMAPPING);
    CheckTable(TT_TAG_GLYPHDATA);
    CheckTable(TT_TAG_FONTHEADER);
    CheckTable(TT_TAG_HORIZONTALHEADER);
    CheckTable(TT_TAG_HORIZONTALMETRICS);
    CheckTable(TT_TAG_INDEXTOLOCATION);
    CheckTable(TT_TAG_MAXIMUMPROFILE);
    CheckTable(TT_TAG_NAMINGTABLE);
    CheckTable(TT_TAG_POSTSCRIPTINFORMATION);
    CheckTable(TT_TAG_WINDOWSSPECIFICMETRICS);
    {$IFDEF TESTCHECKSUMS}
    for i := 0 to Pred(n) do
      if not VerifyTableChecksum(i) then
        Raise ETrueTypeError.CreateFmt('Invalid checksum for table %d: ' + AnsiChar(fTableDirectory[i].tag[0]) + AnsiChar(fTableDirectory[i].tag[1]) + AnsiChar(fTableDirectory[i].tag[2]) + AnsiChar(fTableDirectory[i].tag[3]), [i]);
    {$ENDIF}
    ReadFontNamingTable;
  except
    fValid := False;
    Raise;
    end;
end;

procedure TTrueTypeFont.SaveToStream(Stream: TStream);
var i, n: integer;
    sz: LongWord;
begin
  CheckValidity;
  Stream.WriteBuffer(fOffsetTable, Sizeof(fOffsetTable));
  n := TTUShortToInt(fOffsetTable.numTables);
  for i := 0 to Pred(n) do
    Stream.WriteBuffer(fTableDirectory[i], Sizeof(fTableDirectory[i]));
  for i := 0 to Pred(n) do
    begin
    sz := TTUlongToInt(fTableDirectory[i].length);
    Stream.Position := TTULongToInt(fTableDirectory[i].offset);
    Stream.WriteBuffer(fTables[i]^, sz);
    end;
end;

procedure TTrueTypeFont.LoadFromFile(const FileName: string);
var S: TFileStream;
begin
  S := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(S);
  finally
    S.Free;
    end;
end;

procedure TTrueTypeFont.SaveToFile(const FileName: string);
var S: TFileStream;
begin
  CheckValidity;
  S := TFileStream.Create(FileName, fmCreate or fmShareDenyWrite);
  try
    SaveToStream(S);
  finally
    S.Free;
    end;
end;

function TTrueTypeFont.CalculateTableChecksum(Index: integer): LongWord;
var p: ^TT_ULONG;
    n: integer;
    x: TT_ULONG;
begin
  CheckTableValidity(Index);
  Result := 0;
  n := TTULongToInt(fTableDirectory[Index].length);
  p := fTables[Index];
  while n > Sizeof(TT_ULONG) do
    begin
    Inc(Result, TTULongToInt(p^));
    Inc(p);
    Dec(n, Sizeof(TT_ULONG));
    end;
  if n > 0 then
    begin
    FillChar(x, Sizeof(x), 0);
    Move(p^, x, n);
    Inc(Result, TTULongToInt(x));
    end;
end;

function TTrueTypeFont.VerifyTableChecksum(Index: integer): boolean;
begin
  Result := CalculateTableChecksum(Index) = TTULongToInt(fTableDirectory[Index].checkSum);
end;

procedure TTrueTypeFont.UpdateTableChecksum(Index: integer);
begin
  TTIntToULong(CalculateTableChecksum(Index), fTableDirectory[Index].checkSum);
end;

function TTrueTypeFont.FindTableIndex(Tag: ShortString; StartingIndex: integer): integer;
  function CompareTag(const Tag: ShortString; TableIndex: integer): boolean;
    begin
      Result := CompareMem(@(Tag[1]), @(fTableDirectory[TableIndex].tag), Sizeof(TT_ULONG));
    end;
begin
  CheckValidity;
  Result := -1;
  if Length(Tag) < Sizeof(TT_ULONG) then
    Tag := Tag + StringOfChar(' ', Sizeof(TT_ULONG) - Length(Tag));
  if StartingIndex < 0 then
    StartingIndex := 0;
  while StartingIndex < Length(fTableDirectory) do
    if CompareTag(Tag, StartingIndex) then
      begin
      Result := StartingIndex;
      Break;
      end
    else
      Inc(StartingIndex);
end;

procedure TTrueTypeFont.UpdateTable(Index: integer; const Data; const Size: LongWord);
var OldPos, OldSize, TablePos: LongWord;
    SizeDiff: int64;
    i: integer;
begin
  CheckTableValidity(Index);
  OldSize := TTULongToInt(fTableDirectory[Index].length);
  SizeDiff := int64(Size) - int64(OldSize);
  if SizeDiff > 0 then
    begin
    // Move all following tables
    OldPos := TTULongToInt(fTableDirectory[Index].offset);
    for i := 0 to Pred(Length(fTableDirectory)) do
      begin
      TablePos := TTULongToInt(fTableDirectory[i].offset);
      if TablePos > OldPos then
        TTIntToUlong(TablePos + SizeDiff, fTableDirectory[i].offset);
      end;
    // Remove all table and create a new one
    FreeMem(fTables[Index]);
    GetMem(fTables[Index], Size);
    end
  else if SizeDiff < 0 then
    FillChar(fTables[Index]^, OldSize, 0);
  Move(Data, fTables[Index]^, Size);
  TTIntToULong(Size, fTableDirectory[Index].length);
  UpdateTableChecksum(Index);
end;

procedure TTrueTypeFont.ApplyChanges;
{$IFDEF UPDATECHECKSUMS}
var i: integer;
{$ENDIF}
begin
  CheckValidity;
  ApplyNamingRecordChanges;
  {$IFDEF UPDATECHECKSUMS}
  for i := 0 to Pred(TTUShortToInt(fOffsetTable.numTables)) do
    UpdateTableChecksum(i);
  {$ENDIF}
end;

procedure TTrueTypeFont.CancelChanges;
begin
  CheckValidity;
end;

procedure TTrueTypeFont.ApplyNamingRecordChanges;
var NamingTableIndex: integer;
    Table: PTT_NamingTable;
    TableSize: integer;
begin
  if not fFontNamingChanged then
    Exit;
  NamingTableIndex := FindTableIndex(TT_TAG_NAMINGTABLE);
  BuildFontNamingTable(Table, TableSize);
  try
    UpdateTable(NamingTableIndex, Table^, TableSize);
  finally
    FreeMem(Table);
    end;
  fFontNamingChanged := False;
end;

procedure TTrueTypeFont.ReadFontNamingTable;
var NamingTableIndex, NameRecordCount: integer;
    NamingTable: PTT_NamingTable;
    NamingRecord: PTT_NamingTableNameRecord;
    NameData: Pointer;
    i, sz: integer;
begin
  NamingTableIndex := FindTableIndex(TT_TAG_NAMINGTABLE);
  NamingTable := fTables[NamingTableIndex];
  NameRecordCount := TTUShortToInt(NamingTable^.numNameRecords);
  NamingRecord := Pointer(Integer(NamingTable) + Sizeof(NamingTable^));
  SetLength(fFontNaming, NameRecordCount);
  for i := 0 to Pred(NameRecordCount) do
    begin
    fFontNaming[i].PlatformID := TTUShortToInt(NamingRecord^.platformID);
    fFontNaming[i].EncodingID := TTUShortToInt(NamingRecord^.encodingID);
    fFontNaming[i].LanguageID := TTUShortToInt(NamingRecord^.languageID);
    fFontNaming[i].NameID := TTUShortToInt(NamingRecord^.nameID);
    sz := TTUShortToInt(NamingRecord^.length);
    if sz > 0 then
      begin
      NameData := Pointer(Integer(NamingTable) + TTUShortToInt(NamingTable^.storageOffsetFromTable) + TTUshortToInt(NamingRecord^.stringOffsetFromStorage));
      if IsFontNamingUnicode(fFontNaming[i].PlatformID, fFontNaming[i].EncodingID) then
        fFontNaming[i].Name := TTWideToWide(NameData, sz shr 1)
      else
        fFontNaming[i].Name := AnsiToWide(NameData, sz);
      end
    else
      fFontNaming[i].Name := '';
    Inc(NamingRecord);
    end;
  fFontNamingChanged := False;
end;

procedure TTrueTypeFont.BuildFontNamingTable(out Table: PTT_NamingTable; out TableSize: integer);
var i, n, sz: integer;
    StorageOffset, StringOffset: integer;
    NamingRecord: PTT_NamingTableNameRecord;
    Storage: Pointer;
    s: AnsiString;
    ps: Pointer;
begin
  n := Length(fFontNaming);
  StorageOffset := Sizeof(TT_NamingTable) + n*Sizeof(TT_NamingTableNameRecord);
  TableSize := StorageOffset;
  for i := 0 to Pred(n) do
    begin
    sz := Length(fFontNaming[i].Name);
    if sz > 0 then
      if IsFontNamingUnicode(fFontNaming[i].PlatformID, fFontNaming[i].EncodingID) then
        Inc(TableSize, 2*sz)
      else
        Inc(TableSize, sz);
    end;
  TableSize := (TableSize + 3) and $fffffffc;
  GetMem(Table, TableSize);
  try
    FillChar(Table^, TableSize, 0);
    TTIntToUShort(0, Table^.formatSelector);
    TTIntToUShort(n, Table^.numNameRecords);
    TTIntToUShort(StorageOffset, Table^.storageOffsetFromTable);
    NamingRecord := Pointer(Integer(Table) + Sizeof(TT_NamingTable));
    Storage := Pointer(Integer(Table) + StorageOffset);
    StringOffset := 0;
    for i := 0 to Pred(n) do
      begin
      TTIntToUShort(fFontNaming[i].PlatformID, NamingRecord^.platformID);
      TTIntToUShort(fFontNaming[i].EncodingID, NamingRecord^.encodingID);
      TTIntToUShort(fFontNaming[i].LanguageID, NamingRecord^.languageID);
      TTIntToUShort(fFontNaming[i].NameID, NamingRecord^.nameID);
      TTIntToUShort(StringOffset, NamingRecord^.stringOffsetFromStorage);
      sz := Length(fFontNaming[i].Name);
      if IsFontNamingUnicode(fFontNaming[i].PlatformID, fFontNaming[i].EncodingID) then
        begin
        sz := sz shl 1;
        TTWideToWide(fFontNaming[i].Name, Storage);
        end
      else
        begin
        s := WideToAnsi(fFontNaming[i].Name);
        ps := @(s[1]);
        Move(ps^, Storage^, sz);
        end;
      TTIntToUShort(sz, NamingRecord^.length);
      Inc(NamingRecord);
      Storage := Pointer(Integer(Storage) + sz);
      Inc(StringOffset, sz);
      end;
  except
    FreeMem(Table);
    Raise;
    end;
end;

function TTrueTypeFont.GetAllFontNaming(FontNameID: integer; out Names: TWideStringArray; PreferredPlatform, PreferredEncoding, PreferredLanguage: integer): WideString;
var Found: boolean;
    i, j, n: integer;
    s: WideString;
begin
  Result := '';
  n := 0;
  SetLength(Names, Length(fFontNaming));
  for i := 0 to Pred(Length(fFontNaming)) do
    if fFontNaming[i].NameID = FontNameID then
      if (PreferredPlatform < 0) or (PreferredPlatform = fFontNaming[i].PlatformID) then
        if (PreferredEncoding < 0) or (PreferredEncoding = fFontNaming[i].EncodingID) then
          if (PreferredLanguage < 0) or (PreferredLanguage = fFontNaming[i].LanguageID) then
            begin
            Found := False;
            s := fFontNaming[i].Name;
            for j := 0 to Pred(n) do
              if Names[j] = s then
                begin
                Found := True;
                Break;
                end;
            if not Found then
              begin
              Names[n] := s;
              Inc(n);
              end;
            end;
  SetLength(Names, n);
  if n > 0 then
    Result := Names[0];
end;

function TTrueTypeFont.GetFontNaming(FontNameID, PreferredPlatform, PreferredEncoding, PreferredLanguage: integer): WideString;
var i: integer;
begin
  Result := '';
  for i := 0 to Pred(Length(fFontNaming)) do
    if fFontNaming[i].NameID = FontNameID then
      if (PreferredPlatform < 0) or (PreferredPlatform = fFontNaming[i].PlatformID) then
        if (PreferredEncoding < 0) or (PreferredEncoding = fFontNaming[i].EncodingID) then
          if (PreferredLanguage < 0) or (PreferredLanguage = fFontNaming[i].LanguageID) then
            begin
            Result := fFontNaming[i].Name;
            Break;
            end;
end;

procedure TTrueTypeFont.SetFontNaming(FontNameID: integer; const Value: WideString; PreferredPlatform, PreferredEncoding, PreferredLanguage: integer);
var i: integer;
    Found: boolean;
begin
  Found := False;
  for i := 0 to Pred(Length(fFontNaming)) do
    if fFontNaming[i].NameID = FontNameID then
      if (PreferredPlatform < 0) or (PreferredPlatform = fFontNaming[i].PlatformID) then
        if (PreferredEncoding < 0) or (PreferredEncoding = fFontNaming[i].EncodingID) then
          if (PreferredLanguage < 0) or (PreferredLanguage = fFontNaming[i].LanguageID) then
            if fFontNaming[i].Name <> Value then
              begin
              fFontNamingChanged := True;
              fFontNaming[i].Name := Value;
              Found := True;
              //Break; // I don't want the break here!
              end;
  if not Found then
    begin
    if PreferredPlatform < 0 then
      PreferredPlatform := TT_FONTPLATFORM_MICROSOFT;
    if PreferredEncoding < 0 then
      case PreferredPlatform of
        TT_FONTPLATFORM_APPLEUNICODE: PreferredEncoding := 0;
        TT_FONTPLATFORM_MACINTOSH:    PreferredEncoding := TT_FONTENCODING_MACINTOSH_ROMAN;
        TT_FONTPLATFORM_MICROSOFT:    PreferredEncoding := TT_FONTENCODING_MICROSOFT_UNICODE;
        else
          PreferredEncoding := 0;
        end;
    if PreferredLanguage < 0 then
      case PreferredPlatform of
        TT_FONTPLATFORM_APPLEUNICODE: PreferredLanguage := 0;
        TT_FONTPLATFORM_MACINTOSH:    PreferredLanguage := TT_FONTLANGUAGE_MACINTOSH_ENGLISH;
        TT_FONTPLATFORM_MICROSOFT:    PreferredLanguage := TT_FONTLANGUAGE_MICROSOFT_ENGLISH_US;
        else
          PreferredLanguage := 0;
        end;
    i := Length(fFontNaming);
    SetLength(fFontNaming, Succ(i));
    fFontNaming[i].PlatformID := PreferredPlatform;
    fFontNaming[i].EncodingID := PreferredEncoding;
    fFontNaming[i].LanguageID := PreferredLanguage;
    fFontNaming[i].NameID := FontNameID;
    fFontNaming[i].Name := Value;
    fFontNamingChanged := True;
    end;
end;

function TTrueTypeFont.IsFontNamingUnicode(NamingRecord: PTT_NamingTableNameRecord): boolean;
begin
  Result := IsFontNamingUnicode(TTUShortToInt(NamingRecord^.platformID), TTUShortToInt(NamingRecord^.encodingID));
end;

function TTrueTypeFont.IsFontNamingUnicode(PlatformID, EncodingID: integer): boolean;
begin
  Result :=
    ((PlatformID = TT_FONTPLATFORM_MICROSOFT) and (EncodingID = TT_FONTENCODING_MICROSOFT_UNICODE)) or
    ((PlatformID = TT_FONTPLATFORM_APPLEUNICODE))
    ;
end;

end.
